name = "patas"
version = "2.0.1"
author = "Diego Souza"
author_email = "diegofpsouza+patas@gmail.com"
url = "https://github.com/diegofps/patas"
description = "Parallelize any program and aggregate results from stdout."

